@@ -42,6 +42,7 @@ import email
 import poplib
 import imaplib
 import re
+import select
 
 try:
     # do we have a recent pykerberos?
@@ -52,6 +53,124 @@ try:
 except ImportError:
     pass
 
+# hashlib only present in python2.5, ssl in python2.6; used together
+# in SSL functionality below
+try:
+    import ssl
+except ImportError:
+    ssl = None
+try:
+    import hashlib
+except ImportError:
+    hashlib = None
+
+# If we have an ssl module:
+if ssl:
+    # Is it recent enough to have hostname matching (Python 3.2+)?
+    try:
+        ssl_match_hostname = ssl.match_hostname
+    except AttributeError:
+    # Running a Python with no hostname matching
+        def _dnsname_match(dn, hostname, max_wildcards=1):
+            """Matching according to RFC 6125, section 6.4.3
+            http://tools.ietf.org/html/rfc6125#section-6.4.3
+            """
+            pats = []
+            if not dn:
+                return False
+        
+            parts = dn.split(r'.')
+            leftmost = parts[0]
+            remainder = parts[1:]
+        
+            wildcards = leftmost.count('*')
+            if wildcards > max_wildcards:
+                # Issue #17980: avoid denials of service by refusing more
+                # than one wildcard per fragment.  A survery of established
+                # policy among SSL implementations showed it to be a
+                # reasonable choice.
+                raise getmailOperationError(
+                    "too many wildcards in certificate DNS name: " + repr(dn))
+        
+            # speed up common case w/o wildcards
+            if not wildcards:
+                return dn.lower() == hostname.lower()
+        
+            # RFC 6125, section 6.4.3, subitem 1.
+            # The client SHOULD NOT attempt to match a presented identifier
+            # in which the wildcard character comprises a label other than
+            # the left-most label.
+            if leftmost == '*':
+                # When '*' is a fragment by itself, it matches a non-empty
+                # dotless fragment.
+                pats.append('[^.]+')
+            elif leftmost.startswith('xn--') or hostname.startswith('xn--'):
+                # RFC 6125, section 6.4.3, subitem 3.
+                # The client SHOULD NOT attempt to match a presented identifier
+                # where the wildcard character is embedded within an A-label or
+                # U-label of an internationalized domain name.
+                pats.append(re.escape(leftmost))
+            else:
+                # Otherwise, '*' matches any dotless string, e.g. www*
+                pats.append(re.escape(leftmost).replace(r'\*', '[^.]*'))
+        
+            # add the remaining fragments, ignore any wildcards
+            for frag in remainder:
+                pats.append(re.escape(frag))
+        
+            pat = re.compile(r'\A' + r'\.'.join(pats) + r'\Z', re.IGNORECASE)
+            return pat.match(hostname)
+        
+        
+        def ssl_match_hostname(cert, hostname):
+            """Verify that *cert* (in decoded format as returned by
+            SSLSocket.getpeercert()) matches the *hostname*. RFC 2818 and
+            RFC 6125 rules are followed, but IP addresses are not accepted
+            for *hostname*.
+        
+            getmailOperationError is raised on failure. On success, the function
+            returns nothing.
+            """
+            if not cert:
+                raise ValueError("empty or no certificate, ssl_match_hostname "
+                                 "needs an SSL socket or SSL context with "
+                                 "either CERT_OPTIONAL or CERT_REQUIRED")
+            dnsnames = []
+            san = cert.get('subjectAltName', ())
+            for key, value in san:
+                if key == 'DNS':
+                    if _dnsname_match(value, hostname):
+                        return
+                    dnsnames.append(value)
+            if not dnsnames:
+                # The subject is only checked when there is no dNSName entry
+                # in subjectAltName
+                for sub in cert.get('subject', ()):
+                    for key, value in sub:
+                        # XXX according to RFC 2818, the most specific
+                        # Common Name must be used.
+                        if key == 'commonName':
+                            if _dnsname_match(value, hostname):
+                                return
+                            dnsnames.append(value)
+            if len(dnsnames) > 1:
+                raise getmailOperationError("hostname %s "
+                    "doesn't match either of %s"
+                    % (hostname, ', '.join(map(repr, dnsnames))))
+            elif len(dnsnames) == 1:
+                raise getmailOperationError("hostname %s "
+                    "doesn't match %s"
+                    % (hostname, dnsnames[0]))
+            else:
+                raise getmailOperationError("no appropriate commonName or "
+                    "subjectAltName fields were found")
+        
+try:
+    from email.header import decode_header
+except ImportError, o:
+    # python < 2.5
+    from email.Header import decode_header
+
 from getmailcore.compatibility import *
 from getmailcore.exceptions import *
 from getmailcore.constants import *
@@ -111,6 +230,18 @@ EAI_NODATA = getattr(socket, 'EAI_NODATA', NO_OBJ)
 EAI_FAIL = getattr(socket, 'EAI_FAIL', NO_OBJ)
 
 
+# Constant for POPSSL
+POP3_SSL_PORT = 995
+
+
+# Python added poplib._MAXLINE somewhere along the way.  As far as I can
+# see, it serves no purpose except to introduce bugs into any software
+# using poplib.  Any computer running Python will have at least some megabytes
+# of userspace memory; arbitrarily causing message retrieval to break if any
+# "line" exceeds 2048 bytes is absolutely stupid.
+poplib._MAXLINE = 1 << 20   # 1MB; decrease this if you're running on a VIC-20
+
+
 #
 # Mix-in classes
 #
@@ -119,6 +250,7 @@ EAI_FAIL = getattr(socket, 'EAI_FAIL', NO_OBJ)
 class POP3initMixIn(object):
     '''Mix-In class to do POP3 non-SSL initialization.
     '''
+    SSL = False
     def _connect(self):
         self.log.trace()
         try:
@@ -139,11 +271,61 @@ class POP3initMixIn(object):
                        + os.linesep)
 
 
+#######################################
+class POP3_SSL_EXTENDED(poplib.POP3_SSL):
+    # Extended SSL support for POP3 (certificate checking, 
+    # fingerprint matching, cipher selection, etc.)
+
+    def __init__(self, host, port=POP3_SSL_PORT, keyfile=None,
+                 certfile=None, ssl_version=None, ca_certs=None,
+                 ssl_ciphers=None):
+        self.host = host
+        self.port = port
+        self.keyfile = keyfile
+        self.certfile = certfile
+        self.ssl_version = ssl_version
+        self.ca_certs = ca_certs
+        self.ssl_ciphers = ssl_ciphers
+
+        self.buffer = ''
+        msg = "getaddrinfo returns an empty list"
+        self.sock = None
+        for res in socket.getaddrinfo(self.host, self.port, 0,
+                                      socket.SOCK_STREAM):
+            (af, socktype, proto, canonname, sa) = res
+            try:
+                self.sock = socket.socket(af, socktype, proto)
+                self.sock.connect(sa)
+            except socket.error, msg:
+                if self.sock:
+                    self.sock.close()
+                self.sock = None
+                continue
+            break
+        if not self.sock:
+            raise socket.error(msg)
+        extra_args = {}
+        if self.ssl_version:
+            extra_args['ssl_version'] = self.ssl_version
+        if self.ca_certs:
+            extra_args['cert_reqs'] = ssl.CERT_REQUIRED
+            extra_args['ca_certs'] = self.ca_certs
+        if self.ssl_ciphers:
+            extra_args['ciphers'] = self.ssl_ciphers
+
+        self.file = self.sock.makefile('rb')
+        self.sslobj = ssl.wrap_socket(self.sock, self.keyfile,
+                                      self.certfile, **extra_args)
+        self._debugging = 0
+        self.welcome = self._getresp()
+
+
 #######################################
 class Py24POP3SSLinitMixIn(object):
     '''Mix-In class to do POP3 over SSL initialization with Python 2.4's
     poplib.POP3_SSL class.
     '''
+    SSL = True
     def _connect(self):
         self.log.trace()
         if not hasattr(socket, 'ssl'):
@@ -151,8 +333,39 @@ class Py24POP3SSLinitMixIn(object):
                 'SSL not supported by this installation of Python'
             )
         (keyfile, certfile) = check_ssl_key_and_cert(self.conf)
+        ca_certs = check_ca_certs(self.conf)
+        ssl_version = check_ssl_version(self.conf)
+        ssl_fingerprints = check_ssl_fingerprints(self.conf)
+        ssl_ciphers = check_ssl_ciphers(self.conf)
+        using_extended_certs_interface = False
         try:
-            if keyfile:
+            if ca_certs or ssl_version or ssl_ciphers:
+                using_extended_certs_interface = True
+                # Python 2.6 or higher required, use above class instead of
+                # vanilla stdlib one
+                msg = ''
+                if keyfile:
+                    msg += 'with keyfile %s, certfile %s' % (keyfile, certfile)
+                if ssl_version:
+                    if msg:
+                        msg += ', '
+                    msg += ('using protocol version %s'
+                            % self.conf['ssl_version'].upper())
+                if ca_certs:
+                    if msg:
+                        msg += ', '
+                    msg += 'with ca_certs %s' % ca_certs
+
+                self.log.trace(
+                    'establishing POP3 SSL connection to %s:%d %s'
+                    % (self.conf['server'], self.conf['port'], msg)
+                    + os.linesep
+                )
+                self.conn = POP3_SSL_EXTENDED(
+                    self.conf['server'], self.conf['port'], keyfile, certfile,
+                    ssl_version, ca_certs, ssl_ciphers
+                )
+            elif keyfile:
                 self.log.trace(
                     'establishing POP3 SSL connection to %s:%d with '
                     'keyfile %s, certfile %s'
@@ -170,6 +383,44 @@ class Py24POP3SSLinitMixIn(object):
                 self.conn = poplib.POP3_SSL(self.conf['server'],
                                             self.conf['port'])
             self.setup_received(self.conn.sock)
+            if ssl and hashlib:
+                sslobj = self.conn.sslobj
+                peercert = sslobj.getpeercert(True)
+                ssl_cipher = sslobj.cipher()
+                if ssl_cipher:
+                    ssl_cipher = '%s:%s:%s' % ssl_cipher
+                if not peercert:
+                    actual_hash = None
+                else:
+                    actual_hash = hashlib.sha256(peercert).hexdigest().lower()
+            else:
+                actual_hash = None
+                ssl_cipher = None
+
+            # Ensure cert is for server we're connecting to
+            if ssl and self.conf['ca_certs']:
+                ssl_match_hostname(
+                    self.conn.sslobj.getpeercert(),
+                    self.conf.get('ssl_cert_hostname', None) 
+                        or self.conf['server']
+                )
+
+            if ssl_fingerprints:
+                if not actual_hash:
+                    raise getmailOperationError(
+                        'socket ssl_fingerprints mismatch (no cert provided)'
+                    )
+
+                any_matches = False
+                for expected_hash in ssl_fingerprints:
+                    if expected_hash == actual_hash:
+                        any_matches = True
+                if not any_matches:
+                    raise getmailOperationError(
+                        'socket ssl_fingerprints mismatch (got %s)'
+                        % actual_hash
+                    )
+
         except poplib.error_proto, o:
             raise getmailOperationError('POP error (%s)' % o)
         except socket.timeout:
@@ -182,8 +433,19 @@ class Py24POP3SSLinitMixIn(object):
             )
 
         self.conn.sock.setblocking(1)
-        self.log.trace('POP3 connection %s established' % self.conn
-                       + os.linesep)
+
+        fingerprint_message = ('POP3 SSL connection %s established'
+                               % self.conn)
+        if actual_hash:
+            fingerprint_message += ' with fingerprint %s' % actual_hash
+        if ssl_cipher:
+            fingerprint_message += ' using cipher %s' % ssl_cipher
+        fingerprint_message += os.linesep
+
+        if self.app_options.get('fingerprint', False):
+            self.log.info(fingerprint_message)
+        else:
+            self.log.trace(fingerprint_message)
 
 
 #######################################
@@ -191,6 +453,7 @@ class Py23POP3SSLinitMixIn(object):
     '''Mix-In class to do POP3 over SSL initialization with custom-implemented
     code to support SSL with Python 2.3's poplib.POP3 class.
     '''
+    SSL = True
     def _connect(self):
         self.log.trace()
         if not hasattr(socket, 'ssl'):
@@ -198,6 +461,15 @@ class Py23POP3SSLinitMixIn(object):
                 'SSL not supported by this installation of Python'
             )
         (keyfile, certfile) = check_ssl_key_and_cert(self.conf)
+        ca_certs = check_ca_certs(self.conf)
+        ssl_version = check_ssl_version(self.conf)
+        ssl_fingerprints = check_ssl_fingerprints(self.conf)
+        ssl_ciphers = check_ssl_ciphers(self.conf)
+        if ca_certs or ssl_version or ssl_ciphers or ssl_fingerprints:
+            raise getmailConfigurationError(
+                'SSL extended options are not supported by this'
+                 ' installation of Python'
+            )
         try:
             if keyfile:
                 self.log.trace(
@@ -216,6 +488,7 @@ class Py23POP3SSLinitMixIn(object):
                     + os.linesep
                 )
                 self.conn = POP3SSL(self.conf['server'], self.conf['port'])
+
             self.setup_received(self.conn.rawsock)
         except poplib.error_proto, o:
             raise getmailOperationError('POP error (%s)' % o)
@@ -237,6 +510,7 @@ class Py23POP3SSLinitMixIn(object):
 class IMAPinitMixIn(object):
     '''Mix-In class to do IMAP non-SSL initialization.
     '''
+    SSL = False
     def _connect(self):
         self.log.trace()
         try:
@@ -254,11 +528,41 @@ class IMAPinitMixIn(object):
                        + os.linesep)
 
 
+#######################################
+class IMAP4_SSL_EXTENDED(imaplib.IMAP4_SSL):
+    # Similar to above, but with extended support for SSL certificate checking,
+    # fingerprints, etc.
+    def __init__(self, host='', port=imaplib.IMAP4_SSL_PORT, keyfile=None, 
+                 certfile=None, ssl_version=None, ca_certs=None, 
+                 ssl_ciphers=None):
+       self.ssl_version = ssl_version
+       self.ca_certs = ca_certs
+       self.ssl_ciphers = ssl_ciphers
+       imaplib.IMAP4_SSL.__init__(self, host, port, keyfile, certfile)
+
+    def open(self, host='', port=imaplib.IMAP4_SSL_PORT):
+       self.host = host
+       self.port = port
+       self.sock = socket.create_connection((host, port))
+       extra_args = {}
+       if self.ssl_version:
+           extra_args['ssl_version'] = self.ssl_version
+       if self.ca_certs:
+           extra_args['cert_reqs'] = ssl.CERT_REQUIRED
+           extra_args['ca_certs'] = self.ca_certs
+       if self.ssl_ciphers:
+           extra_args['ciphers'] = self.ssl_ciphers
+
+       self.sslobj = ssl.wrap_socket(self.sock, self.keyfile, self.certfile, 
+                                     **extra_args)
+       self.file = self.sslobj.makefile('rb')
+
 
 #######################################
 class IMAPSSLinitMixIn(object):
     '''Mix-In class to do IMAP over SSL initialization.
     '''
+    SSL = True
     def _connect(self):
         self.log.trace()
         if not hasattr(socket, 'ssl'):
@@ -266,8 +570,39 @@ class IMAPSSLinitMixIn(object):
                 'SSL not supported by this installation of Python'
             )
         (keyfile, certfile) = check_ssl_key_and_cert(self.conf)
+        ca_certs = check_ca_certs(self.conf)
+        ssl_version = check_ssl_version(self.conf)
+        ssl_fingerprints = check_ssl_fingerprints(self.conf)
+        ssl_ciphers = check_ssl_ciphers(self.conf)
+        using_extended_certs_interface = False
         try:
-            if keyfile:
+            if ca_certs or ssl_version or ssl_ciphers:
+                using_extended_certs_interface = True
+                # Python 2.6 or higher required, use above class instead of
+                # vanilla stdlib one
+                msg = ''
+                if keyfile:
+                    msg += 'with keyfile %s, certfile %s' % (keyfile, certfile)
+                if ssl_version:
+                    if msg:
+                        msg += ', '
+                    msg += ('using protocol version %s' 
+                            % self.conf['ssl_version'].upper())
+                if ca_certs:
+                    if msg:
+                        msg += ', '
+                    msg += 'with ca_certs %s' % ca_certs
+
+                self.log.trace(
+                    'establishing IMAP SSL connection to %s:%d %s'
+                    % (self.conf['server'], self.conf['port'], msg)
+                    + os.linesep
+                )
+                self.conn = IMAP4_SSL_EXTENDED(
+                    self.conf['server'], self.conf['port'], keyfile, certfile, 
+                    ssl_version, ca_certs, ssl_ciphers
+                )
+            elif keyfile:
                 self.log.trace(
                     'establishing IMAP SSL connection to %s:%d with keyfile '
                     '%s, certfile %s'
@@ -286,6 +621,44 @@ class IMAPSSLinitMixIn(object):
                 self.conn = imaplib.IMAP4_SSL(self.conf['server'],
                                               self.conf['port'])
             self.setup_received(self.conn.sock)
+            if ssl and hashlib:
+                sslobj = self.conn.ssl()
+                peercert = sslobj.getpeercert(True)
+                ssl_cipher = sslobj.cipher()
+                if ssl_cipher:
+                    ssl_cipher = '%s:%s:%s' % ssl_cipher
+                if not peercert:
+                    actual_hash = None
+                else:
+                    actual_hash = hashlib.sha256(peercert).hexdigest().lower()
+            else:
+                actual_hash = None
+                ssl_cipher = None
+
+            # Ensure cert is for server we're connecting to
+            if ssl and self.conf['ca_certs']:
+                ssl_match_hostname(
+                    self.conn.ssl().getpeercert(),
+                    self.conf.get('ssl_cert_hostname', None) 
+                        or self.conf['server']
+                )
+
+            if ssl_fingerprints:
+                if not actual_hash:
+                    raise getmailOperationError(
+                        'socket ssl_fingerprints mismatch (no cert provided)'
+                    )
+
+                any_matches = False
+                for expected_hash in ssl_fingerprints:
+                    if expected_hash == actual_hash:
+                        any_matches = True
+                if not any_matches:
+                    raise getmailOperationError(
+                        'socket ssl_fingerprints mismatch (got %s)' 
+                        % actual_hash
+                    )
+
         except imaplib.IMAP4.error, o:
             raise getmailOperationError('IMAP error (%s)' % o)
         except socket.timeout:
@@ -305,14 +678,25 @@ class IMAPSSLinitMixIn(object):
                     % (self.conf['server'], o)
                 )
             else:
-                raise getmailOperationError('socket error during connect (%s)' % o)
+                raise getmailOperationError('socket error during connect (%s)' 
+                                            % o)
         except socket.sslerror, o:
             raise getmailOperationError(
                 'socket sslerror during connect (%s)' % o
             )
 
-        self.log.trace('IMAP SSL connection %s established' % self.conn
-                       + os.linesep)
+        fingerprint_message = ('IMAP SSL connection %s established'
+                               % self.conn)
+        if actual_hash:
+            fingerprint_message += ' with fingerprint %s' % actual_hash
+        if ssl_cipher:
+            fingerprint_message += ' using cipher %s' % ssl_cipher
+        fingerprint_message += os.linesep
+
+        if self.app_options['fingerprint']:
+            self.log.info(fingerprint_message)
+        else:
+            self.log.trace(fingerprint_message)
 
 #
 # Base classes
@@ -372,18 +756,21 @@ class RetrieverSkeleton(ConfigurableBase):
       initialize(self, options)
       checkconf(self)
     '''
-
     def __init__(self, **args):
         self.headercache = {}
         self.deleted = {}
-        self.timestamp = int(time.time())
+        self.set_new_timestamp()
         self.__oldmail_written = False
         self.__initialized = False
         self.gotmsglist = False
         self._clear_state()
         self.conn = None
+        self.supports_idle = False
         ConfigurableBase.__init__(self, **args)
 
+    def set_new_timestamp(self):
+        self.timestamp = int(time.time())
+
     def _clear_state(self):
         self.msgnum_by_msgid = {}
         self.msgid_by_msgnum = {}
@@ -391,6 +778,7 @@ class RetrieverSkeleton(ConfigurableBase):
         self.msgsizes = {}
         self.oldmail = {}
         self.__delivered = {}
+        self.deleted = {}
         self.mailbox_selected = False
         
     def setup_received(self, sock):
@@ -403,7 +791,7 @@ class RetrieverSkeleton(ConfigurableBase):
             self.remoteaddr = '[%s]:%s' % serveraddr[:2]
         else:
             # Shouldn't happen
-            self.log.warn('unexpected peer address format %s', str(serveraddr))
+            self.log.warning('unexpected peer address format %s' % str(serveraddr))
             self.remoteaddr = str(serveraddr)
         self.received_from = '%s (%s)' % (self.conf['server'], 
                                           self.remoteaddr)
@@ -728,10 +1116,28 @@ class POP3RetrieverBase(RetrieverSkeleton):
         self.mailboxes = (None, )
         # Handle password
         if self.conf.get('password', None) is None:
-            self.conf['password'] = get_password(
-                self, self.conf['username'], self.conf['server'],
-                self.received_with, self.log
-            )
+            if self.conf.get('password_command', None):
+                # Retrieve from an arbitrary external command
+                command = self.conf['password_command'][0]
+                args = self.conf['password_command'][1:]
+                (rc, stdout, stderr) = run_command(command, args)
+                if stderr:
+                    self.log.warning(
+                        'External password program "%s" wrote to stderr: %s'
+                        % (command, stderr)
+                    )
+                if rc:
+                    # program exited nonzero
+                    raise getmailOperationError(
+                        'External password program error (exited %d)' % rc
+                    )
+                else:
+                    self.conf['password'] = stdout
+            else:
+                self.conf['password'] = get_password(
+                    self, self.conf['username'], self.conf['server'],
+                    self.received_with, self.log
+                )
         RetrieverSkeleton.initialize(self, options)
         try:
             self._connect()
@@ -801,7 +1207,8 @@ class MultidropPOP3RetrieverBase(POP3RetrieverBase):
         if self.envrecipname in NOT_ENVELOPE_RECIPIENT_HEADERS:
             raise getmailConfigurationError(
                 'the %s header field does not record the envelope '
-                'recipient address'
+                    'recipient address'
+                % self.envrecipname
             )
         self.envrecipnum = 0
         try:
@@ -967,9 +1374,10 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                     raise ValueError
                 name = parts.pop(0).lower()
                 r[name] = parts.pop(0)
-        except (ValueError, IndexError), o:
+        except (ValueError, IndexError, AttributeError), o:
             raise getmailOperationError(
-                'IMAP error (failed to parse UID response line "%s")' % line
+                'IMAP error (failed to parse attr response line "%s": %s)' 
+                % (line, o)
             )
         self.log.trace('got %s' % r + os.linesep)
         return r
@@ -1000,18 +1408,34 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                                             % g['mailbox'])
         return mailboxes
 
+    def close_mailbox(self):
+        # Close current mailbox so deleted mail is expunged.  One getmail
+        # user had a buggy IMAP server that didn't do the automatic expunge,
+        # so we do it explicitly here if we've deleted any messages.
+        if self.deleted:
+            self.conn.expunge()
+        self.conn.close()
+        self.write_oldmailfile(self.mailbox_selected)
+        # And clear some state
+        self.mailbox_selected = False
+        self.mailbox = None
+        self.uidvalidity = None
+        self.msgnum_by_msgid = {}
+        self.msgid_by_msgnum = {}
+        self.sorted_msgnum_msgid = ()
+        self._mboxuids = {}
+        self._mboxuidorder = []
+        self.msgsizes = {}
+        self.oldmail = {}
+        self.__delivered = {}
+
     def select_mailbox(self, mailbox):
         self.log.trace()
         assert mailbox in self.mailboxes, (
             'mailbox not in config (%s)' % mailbox
         )
         if self.mailbox_selected is not False:
-            # Close current mailbox so deleted mail is expunged.
-            # Except one user reports that an explicit expunge is needed with 
-            # his server, or else the mail is never removed from the mailbox.
-            self.conn.expunge()
-            self.conn.close()
-            self.write_oldmailfile(self.mailbox_selected)
+            self.close_mailbox()
 
         self._clear_state()
 
@@ -1060,6 +1484,10 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                     'FETCH', '1:%d' % msgcount, '(UID RFC822.SIZE)'
                 )
                 for line in response:
+                    if not line:
+                        # One user had a server that returned a null response
+                        # somehow -- try to just skip.
+                        continue
                     r = self._parse_imapattrresponse(line)
                     # Don't allow / in UIDs we store, as we look for that to 
                     # detect old-style oldmail files.  Can occur with IMAP, at 
@@ -1104,7 +1532,7 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                 )
             self.log.debug('deleting message "%s"' % uid + os.linesep)
             response = self._parse_imapuidcmdresponse(
-                'STORE', uid, 'FLAGS', '(\Deleted)'
+                'STORE', uid, 'FLAGS', '(\Deleted \Seen)'
             )
         except imaplib.IMAP4.error, o:
             raise getmailOperationError('IMAP error (%s)' % o)
@@ -1154,8 +1582,9 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                                             % msgid)
 
             # record mailbox retrieved from in a header
-            msg.add_header('X-getmail-retrieved-from-mailbox', 
-                           self.mailbox_selected)
+            if self.conf['record_mailbox']:
+                msg.add_header('X-getmail-retrieved-from-mailbox', 
+                               self.mailbox_selected)
 
             # google extensions: apply labels, etc
             if 'X-GM-EXT-1' in self.conn.capabilities:
@@ -1185,7 +1614,7 @@ class IMAPRetrieverBase(RetrieverSkeleton):
             self.log.warning('Could not fetch google imap extensions: %s' % o)
             return {}
 
-        if not response:
+        if not response or not response[0]:
             return {}
             
         ext = re.search(
@@ -1230,10 +1659,28 @@ class IMAPRetrieverBase(RetrieverSkeleton):
         # Handle password
         if (self.conf.get('password', None) is None
                 and not (HAVE_KERBEROS_GSS and self.conf['use_kerberos'])):
-            self.conf['password'] = get_password(
-                self, self.conf['username'], self.conf['server'], 
-                self.received_with, self.log
-            )
+            if self.conf['password_command']:
+                # Retrieve from an arbitrary external command
+                command = self.conf['password_command'][0]
+                args = self.conf['password_command'][1:]
+                (rc, stdout, stderr) = run_command(command, args)
+                if stderr:
+                    self.log.warning(
+                        'External password program "%s" wrote to stderr: %s'
+                        % (command, stderr)
+                    )
+                if rc:
+                    # program exited nonzero
+                    raise getmailOperationError(
+                        'External password program error (exited %d)' % rc
+                    )
+                else:
+                    self.conf['password'] = stdout
+            else:
+                self.conf['password'] = get_password(
+                    self, self.conf['username'], self.conf['server'], 
+                    self.received_with, self.log
+                )
             
         RetrieverSkeleton.initialize(self, options)
         try:
@@ -1248,13 +1695,21 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                         'login_cram_md5', self.conf['username'],
                         self.conf['password']
                     )
+                elif self.conf['use_xoauth2']:
+                    # octal 1 / ctrl-A used as separator
+                    auth = 'user=%s\1auth=Bearer %s\1\1' % (self.conf['username'],
+                                                            self.conf['password'])
+                    self.conn.authenticate('XOAUTH2', lambda unused: auth)
                 else:
                     self._parse_imapcmdresponse('login', self.conf['username'],
                                                 self.conf['password'])
             except imaplib.IMAP4.abort, o:
                 raise getmailLoginRefusedError(o)
             except imaplib.IMAP4.error, o:
-                raise getmailCredentialError(o)
+                if o.startswith('[UNAVAILABLE]'):
+                    raise getmailLoginRefusedError(o)
+                else:
+                    raise getmailCredentialError(o)
 
             self.log.trace('logged in' + os.linesep)
             """
@@ -1275,7 +1730,22 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                                    + os.linesep)
                     del self.oldmail[msgid]
             """
-            
+            # Some IMAP servers change the available capabilities after 
+            # authentication, i.e. they present a limited set before login.
+            # The Python stlib IMAP4 class doesn't take this into account
+            # and just checks the capabilities immediately after connecting.
+            # Force a re-check now that we've authenticated.
+            (typ, dat) = self.conn.capability()
+            if dat == [None]:
+                # No response, don't update the stored capabilities
+                self.log.warning('no post-login CAPABILITY response from server\n')
+            else:
+                self.conn.capabilities = tuple(dat[-1].upper().split())
+
+            if 'IDLE' in self.conn.capabilities:
+                self.supports_idle = True
+                imaplib.Commands['IDLE'] = ('AUTH', 'SELECTED')
+
             if self.mailboxes == ('ALL', ):
                 # Special value meaning all mailboxes in account
                 self.mailboxes = tuple(self.list_mailboxes())
@@ -1294,18 +1764,79 @@ class IMAPRetrieverBase(RetrieverSkeleton):
             pass
         self.conn = None
 
+    def go_idle(self, folder, timeout=300):
+        """Initiates IMAP's IDLE mode if the server supports it
+
+        Waits until state of current mailbox changes, and then returns. Returns
+        True if the connection still seems to be up, False otherwise.
+
+        May throw getmailOperationError if the server refuses the IDLE setup
+        (e.g. if the server does not support IDLE)
+
+        Default timeout is 5 minutes.
+        """
+
+        if not self.supports_idle:
+            self.log.warning('IDLE not supported, so not idling\n')
+            raise getmailOperationError(
+                'IMAP4 IDLE requested, but not supported by server'
+            )
+
+
+        if self.SSL:
+            sock = self.conn.ssl()
+        else:
+            sock = self.conn.socket()
+
+        # Based on current imaplib IDLE patch: http://bugs.python.org/issue11245
+        self.conn.untagged_responses = {}
+        self.conn.select(folder)
+        tag = self.conn._command('IDLE')
+        data = self.conn._get_response() # read continuation response
+
+        if data is not None:
+            raise getmailOperationError(
+                'IMAP4 IDLE requested, but server refused IDLE request: %s' 
+                % data
+            )
+
+        self.log.debug('Entering IDLE mode (server says "%s")\n' 
+                       % self.conn.continuation_response)
+
+        try:
+            aborted = None
+            (readable, unused, unused) = select.select([sock], [], [], timeout)
+        except KeyboardInterrupt, o:
+            # Delay raising this until we've stopped IDLE mode
+            aborted = o
+
+        if aborted is not None:
+            self.log.debug('IDLE mode cancelled\n')
+        elif readable:
+            # The socket has data waiting; server has updated status
+            self.log.info('IDLE message received\n')
+        else:
+            self.log.debug('IDLE timeout (%ds)\n' % timeout)
+
+        try:
+            self.conn.untagged_responses = {}
+            self.conn.send('DONE\r\n')
+            self.conn._command_complete('IDLE', tag)
+        except imaplib.IMAP4.error, o:
+            return False
+
+        if aborted:
+            raise aborted
+
+        return True
+
     def quit(self):
         self.log.trace()
         if not self.conn:
             return
         try:
             if self.mailbox_selected is not False:
-                # Close current mailbox so deleted mail is expunged.
-                # Except one user reports that an explicit expunge is needed 
-                # with his server, or else the mail is never removed from the 
-                # mailbox.
-                self.conn.expunge()
-                self.conn.close()
+                self.close_mailbox()
             self.conn.logout()
         except imaplib.IMAP4.error, o:
             #raise getmailOperationError('IMAP error (%s)' % o)
@@ -1335,7 +1866,8 @@ class MultidropIMAPRetrieverBase(IMAPRetrieverBase):
         if self.envrecipname in NOT_ENVELOPE_RECIPIENT_HEADERS:
             raise getmailConfigurationError(
                 'the %s header field does not record the envelope recipient '
-                'address'
+                    'address'
+                % self.envrecipname
             )
         self.envrecipnum = 0
         try:
@@ -1355,13 +1887,14 @@ class MultidropIMAPRetrieverBase(IMAPRetrieverBase):
         self.log.trace()
         msg = IMAPRetrieverBase._getmsgbyid(self, msgid)
         data = {}
-        for (name, val) in msg.headers():
+        for (name, encoded_value) in msg.headers():
             name = name.lower()
-            val = val.strip()
-            if name in data:
-                data[name].append(val)
-            else:
-                data[name] = [val]
+            for (val, encoding) in decode_header(encoded_value):
+                val = val.strip()
+                if name in data:
+                    data[name].append(val)
+                else:
+                    data[name] = [val]
 
         try:
             line = data[self.envrecipname][self.envrecipnum]
